/*
 * Copyright (c) 1990 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include "lber.h"
#include "ldap.h"
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/syslog.h>
#include <sys/wait.h>
#include <signal.h>

#ifndef FD_SET
#define NFDBITS         32
#define FD_SETSIZE      32
#define FD_SET(n, p)    ((p)->fds_bits[(n)/NFDBITS] |= (1 << ((n) % NFDBITS)))
#define FD_CLR(n, p)    ((p)->fds_bits[(n)/NFDBITS] &= ~(1 << ((n) % NFDBITS)))
#define FD_ISSET(n, p)  ((p)->fds_bits[(n)/NFDBITS] & (1 << ((n) % NFDBITS)))
#define FD_ZERO(p)      bzero((char *)(p), sizeof(*(p)))
#endif

#define DAPUSER			"cn=Finger, ou=Miscellaneous Servers, o=University of Michigan, c=US"
#define DEFAULT_BASE		"o=University of Michigan, c=US"
#define DEFAULT_LDAPHOST	"localhost"
#define DEFAULT_PORT		79
#define DEFAULT_SIZELIMIT	50
#define FINGER			"/usr/ucb/finger"

int	debug;
int	dosyslog = 1;

char	*ldaphost = DEFAULT_LDAPHOST;
char	*base = DEFAULT_BASE;
int	strip = 3;
int	searchaliases = 1;
int	deref;
int	sizelimit;

static usage( name )
char	*name;
{
	fprintf( stderr, "usage: %s [-d debuglevel] [-p port] [-l] [-x ldaphost] [-b searchbase] [-a] [-s stripcount] [-z sizelimit\r\n" );
	exit( 1 );
}

main (argc, argv)
int	argc;
char	**argv;
{
	int			s, ns, rc;
	int			port = -1;
	int			tblsize;
	int			i, pid;
	char			*myname;
	unsigned long		mypeer = -1;
	fd_set			readfds;
	struct hostent		*hp;
	struct timeval		timeout;
	struct sockaddr_in	from;
	int			fromlen;
	int			wait4child();
	extern char		*optarg;

	deref = LDAP_DEREF_ALWAYS;
	while ( (i = getopt( argc, argv, "ab:d:lp:s:x:" )) != EOF ) {
		switch( i ) {
		case 'a':	/* do not deref aliases when searching */
			deref = LDAP_DEREF_FINDING;
			break;

		case 'b':	/* search base */
			base = strdup( optarg );
			break;

		case 'd':	/* turn on debugging */
			debug = atoi( optarg );
			break;

		case 'l':	/* do syslogging */
			dosyslog = 1;
			break;

		case 's':	/* number of components to strip off dns */
			strip = atoi( optarg );
			break;

		case 'x':	/* specify ldap host */
			ldaphost = strdup( optarg );
			break;

		case 'z':	/* size limit */
			sizelimit = atoi( optarg );
			break;

		default:
			usage( argv[0] );
		}
	}
	if ( argc > 1 && *argv[1] != '-' ) {
		mypeer = (unsigned long) strtol( argv[1], NULL, 16 );
	}
	printf( "X.500 Finger Service...\r\n" );
	fflush( stdout );

	if ( (myname = strrchr( argv[0], '/' )) == NULL )
		myname = strdup( argv[0] );
	else
		myname = strdup( myname + 1 );

	if ( dosyslog && openlog( myname, LOG_PID | LOG_NOWAIT, LOG_LOCAL4 )
	    < 0 ) {
		perror("openlog");
	}

	if ( dosyslog && mypeer != -1 ) {
		struct in_addr	addr;

		hp = gethostbyaddr( (char *) &mypeer, sizeof(mypeer), AF_INET );
		addr.s_addr = mypeer;
		syslog( LOG_INFO, "connection from %s (%s)", (hp == NULL) ?
		    "unknown" : hp->h_name, inet_ntoa( addr ) );
	}

	do_query();
}

static do_query()
{
	char		buf[256];
	int		len, i;
	int		rc, tblsize;
	struct timeval	timeout;
	fd_set		readfds;
	LDAP		*ld;

	if ( (ld = ldap_open( ldaphost, LDAP_PORT )) == NULL ) {
		perror( "ldap_open" );
		exit( 1 );
	}
	ld->ld_sizelimit = sizelimit ? sizelimit : DEFAULT_SIZELIMIT;
	ld->ld_deref = deref;

	if ( ldap_bind( ld, DAPUSER, NULL, LDAP_AUTH_SIMPLE ) !=
	    LDAP_SUCCESS ) {
		fprintf( stderr, "Sorry, the X.500 service is temporarily unavailable.\n" );
		fprintf( stderr, "Please try again later.\n" );
		ldap_perror( ld, "ldap_bind" );
		exit( 1 );
	}

	tblsize = getdtablesize();
	timeout.tv_sec = 60;
	timeout.tv_usec = 0;
	FD_ZERO( &readfds );
	FD_SET( fileno( stdin ), &readfds );

	if ( (rc = select( tblsize, &readfds, 0, 0, &timeout )) <= 0 ) {
		if ( rc < 0 )
			perror( "select" );
		else
			fprintf( stderr, "connection timed out on input\r\n" );
		exit( 1 );
	}

	if ( fgets( buf, sizeof(buf), stdin ) == NULL )
		exit( 1 );

	len = strlen( buf );
	if ( debug ) {
		fprintf( stderr, "got %d bytes\r\n", len );
#ifdef LDAP_DEBUG
		lber_bprint( buf, len );
#endif
	}

	/* strip off \r \n */
	if ( buf[len - 1] == '\n' ) {
		buf[len - 1] = '\0';
		len--;
	}
	if ( buf[len - 1] == '\r' ) {
		buf[len - 1] = '\0';
		len--;
	}

	/* turn .'s and _'s into spaces */
	for ( i = 0; i < len; i++ ) {
		if ( buf[i] == '.' || buf[i] == '_' || buf[i] == '*' )
			buf[i] = ' ';
	}

	if ( len == 0 ) {
		printf( "No campus-wide login information available.  Info for this machine only:\r\n" );
		fflush( stdout );
		execl( FINGER, FINGER, NULL );
	} else {
		do_search( ld, buf );
	}
}

static do_search( ld, buf )
LDAP	*ld;
char	*buf;
{
	char	filter[1024];
	char	*dn, *rdn;
	char	**title;
	int	rc, matches, spaces, i, approximate = 0;
	struct timeval	tv;
	LDAPMessage	*result, *e;
	static char	*attrs[] = { "cn", "title", 0 };

	spaces = 0;
	for ( i = 0; buf[i] != '\0'; i++ ) {
		if ( isspace( buf[i] ) ) {
			spaces = 1;
			break;
		}
	}

	if ( spaces )
		sprintf( filter, "(|(cn=%s)(sn=%s))", buf, buf );
	else
		sprintf( filter, "(|(cn=%s)(sn=%s)(uid=%s))", buf, buf, buf );

	if ( (rc = ldap_search_s( ld, base, LDAP_SCOPE_SUBTREE, filter, attrs,
	    0, &result )) == -1 ) {
		ldap_perror( ld, "ldap_search" );
		exit( 1 );
	}

	matches = ldap_count_entries( ld, result );
	if ( matches == 0 ) {
		if ( spaces )
			sprintf( filter, "(|(cn~=%s)(sn~=%s))", buf, buf );
		else
			sprintf( filter, "(sn~=%s)", buf );

		if ( (rc = ldap_search_s( ld, base, LDAP_SCOPE_SUBTREE, filter,
		    attrs, 0, &result )) == -1 ) {
			ldap_perror( ld, "ldap_search" );
			exit( 1 );
		}
		approximate = 1;
		matches = ldap_count_entries( ld, result );
	}

	if ( rc == LDAP_SIZELIMIT_EXCEEDED )
		printf( "(Partial results only - a size limit was exceeded)\r\n" );
	else if ( rc == LDAP_TIMELIMIT_EXCEEDED )
		printf( "(Partial results only - a time limit was exceeded)\r\n" );

	if ( matches == 1 ) {
		if ( approximate )
			printf( "One approximate match found for '%s':\r\n",
			    buf );
		e = ldap_first_entry( ld, result );
		do_read( ld, ldap_get_dn( ld, e ) );
	} else if ( matches > 1 ) {
		if ( approximate )
			printf( "%d approximate matches for '%s':\r\n",
			    matches, buf );
		else
			printf( "%d matches for '%s':\r\n", matches, buf );
		fflush( stdout );

		for ( e = ldap_first_entry( ld, result ); e != NULL;
		    e = ldap_next_entry( ld, e ) ) {
			char	*p;

			dn = ldap_get_dn( ld, e );
			rdn = dn;
			if ( (p = strchr( dn, ',' )) != NULL )
				*p = '\0';
			while ( *rdn && *rdn != '=' )
				rdn++;
			if ( *rdn )
				rdn++;
			if ( strcasecmp( rdn, buf ) == 0 ) {
				char	**cn;
				char	*s;
				int	i, last;

				cn = ldap_get_values( ld, e, "cn" );
				for ( i = 0; cn[i] != NULL; i++ ) {
					last = strlen( cn[i] ) - 1;
					if ( isdigit( cn[i][last] ) ) {
						rdn = strdup( cn[i] );
						break;
					}
				}
			}
					
			title = ldap_get_values( ld, e, "title" );

			printf( "  %-20s    %s\r\n", rdn,
			    title ? title[0] : "" );
			fflush( stdout );

			if ( title != NULL )
				ldap_value_free( title );

			free( dn );
		}
	} else if ( matches == 0 ) {
		printf( "Search failed to find anything.  Currently, this service contains information\r\n" );
		printf( "only on University of Michigan faculty, staff and students.  Other likely\r\n" );
		printf( "places to finger are:\r\n" );
		printf( "    um.cc.umich.edu\r\n" );
		printf( "    caen.engin.umich.edu\r\n" );
		printf( "    med.umich.edu\r\n" );
		fflush( stdout );
	} else {
		fprintf( stderr, "error return from ldap_count_entries\r\n" );
		exit( 1 );
	}

	ldap_unbind( ld );
}

static print_attr( ld, fp, label, attr, e, multiline )
LDAP	*ld;
FILE	*fp;
char	*label;
char	*attr;
char	*e;
int	multiline;
{
	char	**val;
	int	i, gotone = 0;

	if ( (val = ldap_get_values( ld, e, attr )) == NULL ) {
		if ( strcmp( attr, "mail" ) == 0 ) {
			fprintf( fp, "%-19s\r\n", label );
			fprintf( fp, "                    None registered in this service.  Try fingering at:\r\n" );
			fprintf( fp, "                    um.cc.umich.edu, caen.engin.umich.edu, med.umich.edu.\r\n" );
		}
		return;
	}

	fprintf( fp, "%-19s\r\n", label );
	for ( i = 0; val[i] != NULL; i++ ) {
		if ( multiline ) {
			char	*s, *p;

			if ( gotone )
				fprintf( fp, "%-19s\r\n", label );
			p = s = val[i];
			while ( s = strchr( s, '$' ) ) {
				*s++ = '\0';
				while ( isspace( *s ) )
					s++;
				fprintf( fp, "                    %s\r\n", p );
				p = s;
			}
			fprintf( fp, "                    %s\r\n", p );
			gotone = 1;
		} else {
			fprintf( fp, "                    %s\r\n", val[i] );
		}
	}
	ldap_value_free( val );
}

static do_read( ld, dn )
LDAP	*ld;
char	*dn;
{
	char		*ufn;
	int		rc, i;
	char		*s;
	char		**cn, **paddr, **tnumber, **uid, **title, **mail;
	struct timeval	tv;
	LDAPMessage	*result, *e;
	static char	*attrs[] = { "cn", "postalAddress", "uid", "title",
				     "mail", "telephoneNumber",
				     "facsimileTelephoneNumber", 
				     "multiLineDescription", 0 };

	if ( (rc = ldap_search_s( ld, dn, LDAP_SCOPE_BASE, "objectclass=*",
	    attrs, 0, &result )) == -1 ) {
		ldap_perror( ld, "ldap_search" );
		exit( 1 );
	}

	if ( ldap_count_entries( ld, result ) != 1 ) {
		printf( "Sorry, an error occurred reading the entry.  Please report this to x500@umich.edu\n" );
		exit( 1 );
	}

	e = ldap_first_entry( ld, result );
	dn = ldap_get_dn( e );
	ufn = ldap_dn2ufn( dn );
	for ( i = 0; i < strip; i++ ) {
		if ( (s = strrchr( ufn, ',' )) == NULL )
			break;
		*s = '\0';
	}
	fprintf( stdout, "\"%s\"\r\n", ufn );
	print_attr( ld, stdout, "  Also known as:", "cn", e, 0 );
	print_attr( ld, stdout, "  E-mail address:", "mail", e, 0 );
	print_attr( ld, stdout, "  Fax number:", "facsimileTelephoneNumber", e, 0 );
	print_attr( ld, stdout, "  Business phone:", "telephoneNumber", e, 0 );
	print_attr( ld, stdout, "  Business address:", "postalAddress", e, 1 );
	print_attr( ld, stdout, "  Title:", "title", e, 0 );
	print_attr( ld, stdout, "  Uniqname:", "uid", e, 0 );
	print_attr( ld, stdout, "  Description:", "multiLineDescription", e, 1 );

	free( ufn );
}
