/*
 * Copyright (c) 1991, 1992 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#ifndef __STDC__
#include <memory.h>
#endif
#include <lber.h>
#include <ldap.h>
#include "ud.h"

char base[BUFSIZ];		/* search base */
extern int verbose;		/* verbose mode flag */
static int num_picked = 0;	/* used when user picks entry at More prompt */
static LDAPMessage *res;	/* results returned from ldap_search_s() */
extern LDAP *ld;		/* our ldap descriptor */
extern int debug;		/* debug flag */

vrfy_the_base(dn)
char *dn;
{
	static char *attrs[2] = { OBJECT_CLASS };

	/* verify that this DN exists in the directory */
	(void) ldap_search_s(ld, dn, LDAP_SCOPE_BASE, "objectClass=*", attrs, TRUE, &res);
	(void) ldap_msgfree(res);
	if ((ld->ld_errno == LDAP_NO_SUCH_OBJECT) || (ld->ld_errno == LDAP_INVALID_DN_SYNTAX))
		return(0);
	else if (ld->ld_errno == LDAP_SUCCESS)
		return(1);
	else {
		ldap_perror(ld, "ldap_search");
		return(0);
	}
}
	

LDAPMessage * find(who)
char *who;
{
	register char *cp;		/* general char pointer */
	register int i;			/* general int */
	int matches;			/* from ldap_count_entries() */
	int choice;			/* entry that user chooses */
	int spaces = 0;			/* if spaces in the name, not # */
	int try_fuzzy = 0;		/* try a fuzzy match */
	static char response[RESP_SIZE];	/* results from user */
	static char *namelist[MAX_NUM_NAMES];	/* names found */
	static char filter[BUFSIZ];		/* search filter */
	static char name[BUFSIZ];		/* DN to lookup */
	static int rc;			/* return from ldap_search */

	register char *rcp = who;		/* for stripping quotes, */
	int seperator = 0;			/* dots, and underscores */

	/* read attributes */
	static char *read_attrs[] = { COMMON_NAME, TITLE, 
		POSTAL_ADDRESS, PHONE_NUMBER, E_MAIL_ADDRESS, 
		HOME_PHONE, HOME_ADDRESS, OBJECT_CLASS, 
		MEMBERS, DESCRIPTION, OWNER, 
		MLDESCRIPTION, FAX, UNIQNAME,
		NULL };

	/* search attributes */
	static char *attrs[] = { COMMON_NAME, TITLE, OBJECT_CLASS, NULL };

	if (debug)
		fprintf(stderr, "find(%s)\n", who);

	/*
	 *  If the name we're looking for has no spaces, treat it
	 *  as a surname, not a commonName.
	 */
	for (cp = who; *cp != '\0'; cp++) {
		if (isspace(*cp)) {
			spaces = 1;
			break;
		}
	}
	
	/*
	 *  This program was modelled after a dreadful program which runs
	 *  on the mainframe here at U-M which requires the user to either
	 *  quote the name, or to separate the parts by dots or underbars.
	 *  We should just strip them because, while we are dreadful, we
	 *  aren't *that* dreadful.
	 */
	for ( ; *rcp != '\0'; rcp++) {
		if ((*rcp == '.') || (*rcp == '_') || (*rcp == '"')) {
			seperator = 1;
			break;
		}
	}
	if (seperator) {
		if (debug)
			fprintf(stderr, "Was: %s\n", who);
		for (rcp = who; *rcp != '\0'; rcp++) {
			if ((*rcp == '.') || (*rcp == '_') || (*rcp == '"')) {
				if (rcp == who)
					who++;
				else if (*(rcp + 1) == '\0')
					*rcp = '\0';
				else {
					*rcp = ' ';
					spaces = 1;
				}
			}
		}
		if (debug)
			fprintf(stderr, "Is: %s\n", who);
	}
			
again:
	/*
	 *  At this point, we could be trying a name for the first time, or
	 *  we could be retrying a name that has failed once already.  If
	 *  we haven't failed (i.e., try_fuzzy is 0), we can approach the
	 *  name as a surname if it has no spaces, or we can treat it as
	 *  a common name if there were spaces.
	 */
	if (!try_fuzzy) {
		if (spaces)
			sprintf(filter, "cn=%s", who);
		else
			sprintf(filter, "(|(sn=%s)(uid=%s))", who, who);
	}
	else {
		if (spaces) {
			char *surname;
			surname = strchr(who, ' ');
			if (surname == NULL)
				sprintf(filter, "cn~=%s", who);
			else
				sprintf(filter, "(|(cn~=%s)(sn=%s))", who, surname);
		}
		else
			sprintf(filter, "sn~=%s", who);
	}

	/*
	 *  Now perform the search and count the results.
	 */
	if ((rc = ldap_search_s(ld, base, LDAP_SCOPE_SUBTREE, filter, attrs,
	    FALSE, &res)) != LDAP_SUCCESS && rc != LDAP_SIZELIMIT_EXCEEDED &&
	    rc != LDAP_SIZELIMIT_EXCEEDED) {
		ldap_perror(ld, "ldap_search_s");
		return(NULL);
	}
	if ((matches = ldap_count_entries(ld, res)) < 0) {
		ldap_perror(ld, "ldap_count_entries");
		fprintf(stderr, "  Buffer contained:\n");
		lber_bprint(res);
		return(NULL);
	}
	else if (matches == 1) {
		if (ldap_search_s(ld, ldap_get_dn(ld, ldap_first_entry(ld, res)), LDAP_SCOPE_BASE, "objectClass=*", read_attrs, FALSE, &res) != LDAP_SUCCESS) {
			ldap_perror(ld, "ldap_search_s");
			return(NULL);
		}
		return(res);
	}
	else if (matches == 0) { /* that didn't work -- try an inexact match */
		/*
		 *  Boo hoo!  No matches.  If we already tried a fuzzy match,
		 *  then we fall through.
		 */
		if (!try_fuzzy) {
			try_fuzzy = 1;
			goto again;
		}
		/*
		 *  If we're here, we're finished unless . . .
		 *
		 *  If we have been trying the name as a surname, and then
		 *  as a fuzzy surname, we still might match it as a common
		 *  name (or a fuzzy common name).
		 */
		if (!strncmp(filter, "sn~=", 4) && !spaces) {
			spaces = 1;		/* treat as a common name */
			try_fuzzy = 0;
			goto again;
		}
		printf(" Could not find \"%s\".\n", who);
		return(NULL);
	}

	/*
	 *  If we are here, it means that we got back multiple answers.
	 *  This can happen with both fuzzy and exact matches.
	 */
	if ((ld->ld_errno == LDAP_TIMELIMIT_EXCEEDED) || (ld->ld_errno == LDAP_SIZELIMIT_EXCEEDED)) {
		if (verbose) {
			printf("  Your query was too general and a limit was exceeded.  The results listed\n");
			printf("  are not complete.  You may want to try again with a more refined query.\n\n");
		}
		else
			printf("  Time or size limit exceeded.  Partial results follow.\n\n");
	}
	printf("  %1d names matched \"%s\".\n", matches, who);
	for (;;) {
		printf("  Do you wish to see a list of names? ");
		fflush(stdout);
		(void) memset(response, 0, sizeof(response));
		fetch_buffer(response, sizeof(response), stdin);
		switch (response[0]) {
		case 'n' :
		case 'N' :
		case '\0' :
		case '\n' :
			return(NULL);
			/* NOTREACHED */
		case 'y' :
		case 'Y' :
			print_list(res, namelist, &matches);
			if (num_picked == 0)
				choice = pick_one(matches);
			else
				choice = num_picked;
			num_picked = 0;
			if (choice >= 0)
				strcpy(name, namelist[choice]);
			/*
			 *  Now free up all of the pointers malloc'ed in
			 *  namelist.
			 */
			for (i = 0; namelist[i] != NULL; i++)
				free(namelist[i]);
			if (choice < 0)
				return(NULL);
			if (ldap_search_s(ld, name, LDAP_SCOPE_BASE, "objectClass=*", read_attrs, FALSE, &res) != LDAP_SUCCESS) {
				ldap_perror(ld, "ldap_search_s");
				return(NULL);
			}
			return(res);
			/* NOTREACHED */
		default :
			printf("  What?\n");
			break;
		}
	}
	/* NOTREACHED */
}

pick_one(i)
int i;
{
	int n;
	char user_pick[RESP_SIZE];

	/* make the user pick an entry */
	for (;;) {
		printf("  Enter the number of the name you want or Q to quit: ");
		fflush(stdout);
		fetch_buffer(user_pick, sizeof(user_pick), stdin);
		if (user_pick[0] == 'q' || user_pick[0] == 'Q')
			return(-1);
		n = atoi(user_pick);
		if ((n > 0) && (n <= i))
			return(n);
		printf("  Invalid response\n");
	}
	/* NOTREACHED */
}

print_list(list, names, matches)
LDAPMessage *list;
char *names[];
int *matches;
{
	char **rdns;
	extern int lpp;
	char resp[RESP_SIZE];
	register LDAPMessage *ep;
	register int i = 1;
	register int rest = 4;		/* 4, not 1 */

	/* print a list of names from which the user will select */
	for (ep = ldap_first_entry(ld, list); ep != NULL; ep = ldap_next_entry(ld, ep)) {
		
		names[i] = ldap_get_dn(ld, ep);
		rdns = ldap_explode_dn(names[i], TRUE);
		printf(" %3d. %s\n", i, *rdns);
		ldap_value_free(rdns);
		i++;
		if ((rest++ > (lpp - 1)) && (i < *matches)) {
again:
			printf("  More? ");
			fflush(stdout);
			fetch_buffer(resp, sizeof(resp), stdin);
			if ((resp[0] == 'n') || (resp[0] == 'N'))
				break;
			else if ((num_picked = atoi(resp)) != 0) {
				if (num_picked < i)
					break;
				else
					goto again;
			}
			rest = 1;
		}
	}
	*matches = i - 1;
	names[i] = NULL;
	return;
}

#ifdef UOFM

char * fetch_update_setting(who)
char *who;
{
	int setting;			/* 1 if TRUE, 0 if FALSE */
	BerElement *cookie;		/* it's magic, you know */
	LDAPMessage *result;		/* from the search below */
	register LDAPMessage *ep;	/* entry pointer */
	register char *ap, **vp;	/* for parsing the result */
	static char *attributes[] = { UPDATES, NULL };

	if (ldap_search_s(ld, who, LDAP_SCOPE_BASE, "objectClass=*", attributes, FALSE, &result) != LDAP_SUCCESS) {
		if (ld->ld_errno == LDAP_NO_SUCH_ATTRIBUTE)
			return("FALSE");
		ldap_perror(ld, "ldap_search_s");
		return(NULL);
	}

	/*
	 *  We did a read on one name and only asked for one attribute.
	 *  There's no reason to loop through any of these structures.
	 *
	 *  We also don't handle the case where noBatchUpdates has more
	 *  than one value.  That should never happen anyhow.
	 *
	 *  If ldap_first_attribute() returns NULL, then this entry did
	 *  not have a noBatchUpdates attribute.
	 */
	ep = ldap_first_entry(ld, result);
	if ((ap = ldap_first_attribute(ld, ep, &cookie)) == NULL)
		return(NULL);
	if (!strcasecmp(ap, UPDATES)) {
		vp = (char **) ldap_get_values(ld, ep, ap);
		if (!strcasecmp(*vp, "TRUE"))
			setting = 1;
		else if (!strcasecmp(*vp, "FALSE"))
			setting = 0;
		else {
			fprintf(stderr, "  Got garbage -> [%s]\n", *vp);
			ldap_value_free(vp);
			return(NULL);
		}
		ldap_value_free(vp);
		if (setting == 1)
			return("TRUE");
		else
			return("FALSE");
	}
	else {
		fprintf(stderr, "  Got garbage -> [%s]\n", ap);
		return(NULL);
	}
}

#endif
